"""
Copyright © SurrealDB Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.

You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

from json import JSONDecodeError
from types import TracebackType
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

import httpx

from ..common import json as jsonlib
from ..common.exceptions import SurrealException
from ..models.response import SurrealResponse

__all__ = ("HTTPClient",)


class HTTPClient:
    """Represents a http connection to a SurrealDB server.

    Parameters
    ----------
    url: :class:`str`
        The URL of the SurrealDB server.
    namespace: :class:`str`
        The namespace to use for the connection.
    database: :class:`str`
        The database to use for the connection.
    username: :class:`str`
        The username to use for the connection.
    password: :class:`str`
        The password to use for the connection.
    """

    def __init__(
        self,
        url: str,
        *,
        namespace: str,
        database: str,
        username: str,
        password: str,
    ) -> None:
        self._url = url
        self._namespace = namespace
        self._database = database
        self._username = username
        self._password = password

        self._http = httpx.AsyncClient(
            base_url=self._url,
            auth=httpx.BasicAuth(
                username=self._username,
                password=self._username,
            ),
            headers={
                "NS": self._namespace,
                "DB": self._database,
                "Accept": "application/json",
                "Content-Type": "application/json",
            },
        )

    async def __aenter__(self) -> HTTPClient:
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_value: Optional[BaseException] = None,
        traceback: Optional[TracebackType] = None,
    ) -> None:
        await self.disconnect()

    async def connect(self) -> None:
        """Connects to http client."""
        await self._http.__aenter__()

    async def disconnect(self) -> None:
        """Disconnects from the http client."""
        await self._http.aclose()

    async def _request(
        self,
        method: str,
        uri: str,
        data: Optional[str] = None,
    ) -> SurrealResponse:
        surreal_response = await self._http.request(
            method=method,
            url=uri,
            content=data,
        )
        surreal_raw_data = await surreal_response.aread()

        try:
            surreal_json: List[Dict[str, Any]] = jsonlib.loads(surreal_raw_data)
            surreal_data = surreal_json[0]
        except JSONDecodeError:
            raise SurrealException(
                f"Invalid JSON response from SurrealDB: {surreal_raw_data}",
            )
        else:
            if surreal_data["status"] != "OK":
                raise SurrealException(
                    f"Query failed: {surreal_data}",
                )

            if surreal_response.status_code not in range(200, 300):
                raise SurrealException(
                    f"Query failed with status code {surreal_response.status_code}: {surreal_data}",
                )

            response_obj = SurrealResponse(**surreal_data)

        return response_obj

    async def execute(self, query: str) -> List[Dict[str, Any]]:
        """Executes a query against the SurrealDB server.

        Parameters
        ----------
        query: :class:`str`
            The query to execute.

        Returns
        -------
        :class:`List[Dict[str, Any]]`
            The results of the query.
        """
        response = await self._request(method="POST", uri="/sql", data=query)
        return response.result

    async def create_all(self, table: str, data: Any) -> List[Dict[str, Any]]:
        """Creates multiple items in a table.

        Parameters
        ----------
        table: :class:`str`
            The table to create the item in.
        data: :class:`Any`
            The data to insert into the table.

        Returns
        -------
        :class:`Dict[str, Any]`
            The items created
        """
        response = await self._request(
            method="POST",
            uri=f"/key/{table}",
            data=jsonlib.dumps(data),
        )

        return response.result

    async def create_one(self, table: str, id: str, data: Any) -> Dict[str, Any]:
        """Creates a single item in a table.

        Parameters
        ----------
        table: :class:`str`
            The table to create the item in.
        id: :class:`str`
            The id of the item to select.
        data: :class:`Any`
            The data to insert into the table.

        Returns
        -------
        :class:`Dict[str, Any]`
            The item created.
        """
        response = await self._request(
            method="POST",
            uri=f"/key/{table}/{id}",
            data=jsonlib.dumps(data),
        )

        return response.result[0]

    async def select_all(self, table: str) -> List[Dict[str, Any]]:
        """Selects all items in a table.

        Parameters
        ----------
        table: :class:`str`
            The table to select from.

        Returns
        -------
        :class:`Dict[str, Any]`
            All items retrieved from the table.
        """
        response = await self._request(method="GET", uri=f"/key/{table}")
        return response.result

    async def select_one(self, table: str, id: str) -> Dict[str, Any]:
        """Selects a single item in a table.

        Parameters
        ----------
        table: :class:`str`
            The table to select from.
        id: :class:`str`
            The id of the item to select.

        Returns
        -------
        :class:`Dict[str, Any]`
            The item selected.
        """
        response = await self._request(method="GET", uri=f"/key/{table}/{id}")
        if not response.result:
            raise SurrealException(f"Key {id} not found in table {table}")

        return response.result[0]

    async def replace_one(self, table: str, id: str, data: Any) -> Dict[str, Any]:
        """Replaces a single item in a table.

        This method requires the entire data structure
        to be sent and will create or update the item.

        Parameters
        ----------
        table: :class:`str`
            The table the item is in.
        id: :class:`str`
            The id of the item to replace.
        data: :class:`Any`
            The data replace the original item with.

        Returns
        -------
        :class:`Dict[str, Any]`
            The new item.
        """
        response = await self._request(
            method="PUT",
            uri=f"/key/{table}/{id}",
            data=jsonlib.dumps(data),
        )
        return response.result[0]

    async def upsert_one(self, table: str, id: str, data: Any) -> Dict[str, Any]:
        """Upserts a single item in a table.

        This method requires only the fields you wish to be updated,
        assuming that this id exists. If it doesn't exist, it will be created
        with the data.

        Parameters
        ----------
        table: :class:`str`
            The table the item is in.
        id: :class:`str`
            The id of the item to upsert.
        data: :class:`Any`
            The data to upsers the original item with.

        Returns
        -------
        :class:`Dict[str, Any]`
            The difference between the old and new item.
        """
        response = await self._request(
            method="PATCH",
            uri=f"/key/{table}/{id}",
            data=jsonlib.dumps(data),
        )

        return response.result[0]

    async def delete_all(self, table: str) -> None:
        """Deletes all items in a table.

        Parameters
        ----------
        table: :class:`str`
            The table to delete all items from.
        """
        await self._request(method="DELETE", uri=f"/key/{table}")

    async def delete_one(self, table: str, id: str) -> None:
        """Deletes one item in a table.

        Parameters
        ----------
        table: :class:`str`
            The table the item is in.
        id: :class:`str`
            The id of the item to delete.
        """
        await self._request(method="DELETE", uri=f"/key/{table}/{id}")



"""
Copyright © SurrealDB Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.

You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

import asyncio
import dataclasses
from types import TracebackType
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

from aiohttp import ClientSession
from aiohttp import ClientWebSocketResponse
from aiohttp import WSMsgType

from ..common import json as jsonlib
from ..common.exceptions import SurrealWebsocketException
from ..common.id import generate_id
from ..models import RPCError
from ..models import RPCRequest
from ..models import RPCResponse

__all__ = ("WebsocketClient",)


class WebsocketClient:
    """Represents a websocket connection
    to a SurrealDB server.

    Parameters
    ----------
    url: :class:`str`
        The URL of the SurrealDB server.
    namespace: :class:`str`
        The namespace to use for the connection.
    database: :class:`str`
        The database to use for the connection.
    username: :class:`str`
        The username to use for the connection.
    password: :class:`str`
        The password to use for the connection.
    """

    def __init__(
        self,
        url: str,
    ) -> None:
        self._url = url

        self._client: ClientSession
        self._ws: ClientWebSocketResponse
        self._recv_task: asyncio.Task

        self._response_futures: Dict[str, asyncio.Future] = {}

    async def __aenter__(self) -> WebsocketClient:
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_value: Optional[BaseException] = None,
        traceback: Optional[TracebackType] = None,
    ) -> None:
        await self.disconnect()

    async def connect(self) -> None:
        """Connects to the SurrealDB server."""
        self._client = ClientSession()
        self._ws = await self._client.ws_connect(self._url)

        self._recv_task = asyncio.create_task(self._receive_task())
        self._recv_task.add_done_callback(self._receive_complete)

    async def disconnect(self) -> None:
        """Disconnects from the SurrealDB server."""
        self._recv_task.cancel()

        await self._ws.close()
        await self._client.close()

    def _receive_complete(self, task: asyncio.Task) -> None:
        try:
            task.result()
        except asyncio.CancelledError:
            pass

    async def _receive_task(self) -> None:
        async for msg in self._ws:
            if msg.type == WSMsgType.ERROR:
                raise SurrealWebsocketException(msg.data)

            json_response: Dict[str, Any] = jsonlib.loads(msg.data)

            error = json_response.get("error")
            if error is not None:
                error = RPCError(**error)
                raise SurrealWebsocketException(error.message)

            response = RPCResponse(**json_response)

            request_future = self._response_futures.pop(response.id, None)
            if request_future is not None:
                request_future.set_result(response)

    async def _send(
        self,
        method: str,
        *params: Any,
    ) -> Any:
        request = RPCRequest(
            id=generate_id(),
            method=method,
            params=params,
        )
        await self._ws.send_json(dataclasses.asdict(request))

        loop = asyncio.get_running_loop()
        future = loop.create_future()
        self._response_futures[request.id] = future

        response: RPCResponse = await asyncio.wait_for(future, timeout=None)
        return response.result

    async def ping(self) -> bool:
        """Pings the SurrealDB server."""
        response = await self._send("ping")
        return response

    async def use(
        self,
        namespace: str,
        database: str,
    ) -> None:
        """Changes the namespace and database to use."""
        response = await self._send("use", namespace, database)

        self._namespace = namespace
        self._database = database

        return response

    async def info(self) -> None:
        response = await self._send("info")
        return response

    async def signup(self, params: Dict[str, Any]) -> str:
        """Creates an account on the SurrealDB server.

        Parameters
        ----------
        params: :class:`Dict[str, Any]`
            The dict of params to pass to sign up.

        Returns
        -------
        :class:`str`
            A JWT token for the new account.
        """
        # if we send None, it's going to error so we must remove any none values
        sanitised_params = {k: v for k, v in params.items() if v is not None}

        response = await self._send("signup", sanitised_params)
        return response

    async def signin(self, params: Dict[str, Any]) -> None:
        """Signs in to the SurrealDB server.

        Parameters
        ----------
        params: :class:`Dict[str, Any]`
            The dict of params to pass to sign in.
        """
        # if we send None, it's going to error so we must remove any none values
        sanitised_params = {k: v for k, v in params.items() if v is not None}

        response = await self._send("signin", sanitised_params)
        return response

    async def invalidate(self) -> None:
        """Invalidates the current session."""
        response = await self._send("invalidate")
        return response

    async def authenticate(self, token: str) -> None:
        """Authenticates the current session.

        Parameters
        ----------
        token: :class:`str`
            The token to use for the connection.
        """
        response = await self._send("authenticate", token)
        return response

    async def live(self, table: str) -> None:
        response = await self._send("live", table)
        return response

    async def kill(self, id: str) -> None:
        response = await self._send("kill", id)
        return response

    async def let(self, key: str, value: Any) -> None:
        """Sets a value in the SurrealDB server.

        Parameters
        ----------
        key: :class:`str`
            The key to set.
        value: Any
            The value to set.
        """
        response = await self._send("let", key, value)
        return response

    async def set(self, key: str, value: Any) -> None:
        """Alias for :meth:`let`."""
        response = await self._send("set", key, value)
        return response

    async def query(self, sql: str, params: Any = None) -> List[List[Any]]:
        """Executes a SQL query.

        Parameters
        ----------
        sql: :class:`str`
            The SQL queries to execute.
        params: Any
            List of parameters which are part of the SQL queries.

        Returns
        -------
        List[List[Any]]
            The results for each query executed.
        """
        response = await self._send("query", sql, params)
        return [query_result["result"] for query_result in response]

    async def select_one(self, table_or_record_id: str) -> Optional[Dict[str, Any]]:
        """Selects a row from an SQL table.

        Parameters
        ----------
        table_or_record_id: :class:`str`
            The table or record ID to find.

        Returns
        -------
        Optional[Dict[:class:`str`, Any]]
            The first row found from the table or record ID, if any.
        """
        response = await self.select(table_or_record_id)
        if not response:
            return None

        return response[0]

    async def select(self, table_or_record_id: str) -> List[Dict[str, Any]]:
        """Selects rows from an SQL table.

        Parameters
        ----------
        table_or_record_id: :class:`str`
            The table or record ID to find.

        Returns
        -------
        List[Dict[:class:`str`, Any]]
            The results of the select.
        """
        response = await self._send("select", table_or_record_id)
        return response

    async def create(
        self,
        table_or_record_id: str,
        data: Any,
    ) -> List[Dict[str, Any]]:
        """Creates a new record in the database.

        Parameters
        ----------
        table_or_record_id: :class:`str`
            The table or record ID to create the record in.
        data: :class:`dict`
            The data to create the record with.

        Returns
        -------
        List[Dict[:class:`str`, Any]]
            The created record.
        """
        response = await self._send("create", table_or_record_id, data)
        return response

    async def update_one(self, record_id: str, data: Any) -> Dict[str, Any]:
        """Updates a record in the database.

        Parameters
        ----------
        record_id: :class:`str`
            The record ID to update the record.
        data: :class:`dict`
            The data to update the record with.

        Returns
        -------
        Dict[:class:`str`, Any]
            The updated record.
        """
        response = await self.update(record_id, data)
        return response[0]

    async def update(
        self,
        table_or_record_id: str,
        data: Any,
    ) -> List[Dict[str, Any]]:
        """Updates a record or records in a table.

        Parameters
        ----------
        table_or_record_id: :class:`str`
            The table or record ID to update.
        data: :class:`dict`
            The data to update the record(s) with.

        Returns
        -------
        List[Dict[:class:`str`, Any]]`
            The updated records.
        """
        response = await self._send("update", table_or_record_id, data)
        return response

    async def change_one(self, record_id: str, data: Any) -> Dict[str, Any]:
        """Alias for :meth:`update_one`."""
        response = await self.change(record_id, data)
        return response[0]

    async def change(
        self,
        table_or_record_id: str,
        data: Any,
    ) -> List[Dict[str, Any]]:
        """Alias for :meth:`update`."""
        response = await self._send("change", table_or_record_id, data)
        return response

    async def modify_one(self, record_id: str, data: Any) -> Dict[str, Any]:
        """Modifies a record or table.

        Parameters
        ----------
        record_id: :class:`str`
            The record ID to modify.
        data: :class:`Any`
            The data to modify the record with.

        Returns
        -------
        Dict[:class:`str`, Any]
            A dictionary containing the diff of the record.
        """
        response = await self.modify(record_id, data)
        return response[0]

    async def modify(
        self,
        table_or_record_id: str,
        data: Any,
    ) -> List[Dict[str, Any]]:
        """Modifies a record or table.

        Parameters
        ----------
        table_or_record_id: :class:`str`
            The table or record ID to modify.
        data: :class:`Any`
            The data to modify the record or table with.

        Returns
        -------
        List[Dict[:class:`str`, Any]]
            A list of dictionaries containing the diff of the records.
        """
        response = await self._send("modify", table_or_record_id, data)
        return response

    async def delete(self, table_or_record_id: str) -> List[Dict[str, Any]]:
        """Deletes a record or table.

        Parameters
        ----------
        table_or_record_id: :class:`str`
            The table or record ID to delete.

        Returns
        -------
        List[Dict[:class:`str`, Any]]
            A list of dictionaries containing the deleted records.
        """
        response = await self._send("delete", table_or_record_id)
        return response



"""
Copyright © SurrealDB Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.

You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Any
from typing import Union


try:
    import orjson as jsonlib
except ImportError:
    import json as jsonlib

__all__ = (
    "dumps",
    "loads",
)


def dumps(obj: Any, **kwargs: Any) -> str:
    result = jsonlib.dumps(obj, **kwargs)
    if isinstance(result, bytes):
        result = result.decode("utf-8")

    return result


def loads(content: Union[bytes, str], **kwargs: Any) -> Any:
    return jsonlib.loads(content, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['surrealdb', 'surrealdb.clients', 'surrealdb.common', 'surrealdb.models']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1', 'httpx>=0.23.0']

setup_kwargs = {
    'name': 'surrealdb',
    'version': '0.1.0',
    'description': 'The official SurrealDB library for Python.',
    'long_description': 'None',
    'author': 'SurrealDB',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
}


setup(**setup_kwargs)
